import unittest

from picograd.engine import Var


class TestEngine(unittest.TestCase):
    def test_scalar(self):
        x = Var(-4.0)
        z = 2 * x + 2 + x
        q = z.relu() + z * x
        h = (z * z).relu()
        y = h + q + q * x
        y.backward()

        # forward pass went well
        self.assertEqual(y.data, -20.0)
        # backward pass went well
        self.assertEqual(x.grad, 46.0)

    def test_scalar_sub(self):
        x = Var(2.5)
        y = Var(3.5)

        z = x - y
        self.assertEqual(z.data, -1)
        z.backward()
        self.assertEqual(x.grad, 1)  # dz/dx
        self.assertEqual(y.grad, -1)  # dz/dy

        # Reset gradients
        x.grad = 0
        y.grad = 0

        z = y - x
        self.assertEqual(z.data, 1)
        z.backward()
        self.assertEqual(x.grad, -1)  # dz/dx
        self.assertEqual(y.grad, 1)  # dz/dy

        # Reset gradient
        x.grad = 0

        z = x - 3
        self.assertEqual(z.data, -0.5)
        z.backward()
        self.assertEqual(x.grad, 1)  # dz/dx

        # Reset gradient
        x.grad = 0

        z = 3 - x
        self.assertEqual(z.data, 0.5)
        z.backward()
        self.assertEqual(x.grad, -1)  # dz/dx

    def test_scalar_div(self):
        x = Var(1.0)
        y = Var(4.0)

        z = x / y
        self.assertEqual(z.data, 0.25)
        z.backward()
        self.assertEqual(x.grad, 0.25)  # dz/dx
        self.assertEqual(y.grad, -0.0625)  # dz/dy

        # Reset gradients
        x.grad = 0
        y.grad = 0

        z = y / x
        self.assertEqual(z.data, 4)
        z.backward()
        self.assertEqual(x.grad, -4)  # dz/dx
        self.assertEqual(y.grad, 1)  # dz/dy

        # Reset gradient
        x.grad = 0

        z = x / 3
        self.assertEqual(z.data, 1 / 3)
        z.backward()
        self.assertEqual(x.grad, 1 / 3)  # dz/dx

        # Reset gradient
        x.grad = 0

        z = 3 / x
        self.assertEqual(z.data, 3)
        z.backward()
        self.assertEqual(x.grad, -3)  # dz/dx

    def test_more_ops(self):
        a = Var(-4.0)
        b = Var(2.0)
        c = a + b
        d = a * b + b ** 3
        c += c + 1
        c += 1 + c + (-a)
        d += d * 2 + (b + a).relu()
        d += 3 * d + (b - a).tanh()
        e = c - d
        f = e ** 2
        g = f / 2.0
        g += 10.0 / f
        h = g.sigmoid()
        h.backward()

        # forward pass went well
        self.assertAlmostEqual(h.data, 0.989, 3)
        # backward pass went well
        self.assertAlmostEqual(a.grad, -0.115, 3)
        self.assertAlmostEqual(b.grad, -0.505, 3)


if __name__ == "__main__":
    unittest.main()



from abc import ABC, abstractmethod
from picograd.engine import Var
from typing import List


class Optimizer(ABC):
    """Base class for optimizers"""

    def __init__(self, parameters: List[Var]) -> None:
        self.parameters: List[Var] = parameters

    def zero_grad(self) -> None:
        """Reset gradients for all parameters"""

        for p in self.parameters:
            p.grad = 0.0

    @abstractmethod
    def step(self) -> None:
        """Take a step of gradient descent"""
        raise NotImplementedError


class SGD(Optimizer):
    """Stochastic Gradient Descent optimizer"""

    def __init__(self, parameters: List[Var], lr: float = 0.01, momentum: float = 0.0, nesterov: bool = False) -> None:
        super(SGD, self).__init__(parameters)
        assert momentum >= 0.0, "momentum cannot be negative"
        self.lr = lr
        self.momentum = momentum
        self.nesterov = nesterov

        self._velocity = [0] * len(parameters)

    def step(self) -> None:
        """Update model parameters in the opposite direction of their gradient"""

        for ind, p in enumerate(self.parameters):
            self._velocity[ind] = (self._velocity[ind] * self.momentum) - self.lr * p.grad
            if self.nesterov:
                p.data += self._velocity[ind] * self.momentum - self.lr * p.grad
            else:
                p.data += self._velocity[ind]


class Adam(Optimizer):
    """Adam optimizer"""

    def __init__(self, parameters: List[Var], lr: float = 1e-3, beta_1: float = 0.9, beta_2: float = 0.999,
                 eps: float = 1e-8) -> None:
        super(Adam, self).__init__(parameters)
        assert (0 <= beta_1) and (beta_1 < 1), "smoothing factor must be in [0,1)"
        assert (0 <= beta_2) and (beta_2 < 1), "smoothing factor must be in [0,1)"

        self.lr = lr
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.eps = eps

        self._t = 0
        self._exp_avg = [0] * len(parameters)
        self._exp_avg_sq = [0] * len(parameters)

    def step(self) -> None:
        self._t += 1

        for ind, p in enumerate(self.parameters):
            self._exp_avg[ind] = self.beta_1 * self._exp_avg[ind] + (1. - self.beta_1) * p.grad
            self._exp_avg_sq[ind] = self.beta_2 * self._exp_avg_sq[ind] + (1. - self.beta_2) * (p.grad ** 2)

            bias_correction_1 = self._exp_avg[ind] / (1. - (self.beta_1 ** self._t))
            bias_correction_2 = self._exp_avg_sq[ind] / (1. - (self.beta_2 ** self._t))

            p.data -= self.lr * bias_correction_1 / (bias_correction_2 ** 0.5 + self.eps)



import unittest

from picograd.engine import Var
from picograd.nn import Neuron, Layer, MLP


class TestNN(unittest.TestCase):
    def test_neuron(self):
        n_features = 3
        neuron = Neuron(in_features=n_features)

        # Check parameter count
        self.assertEqual(len(neuron.parameters()), n_features + 1)  # Including bias

        # Overwite parameters for output prediction
        neuron.w = list(Var(x) for x in range(n_features))
        neuron.b = Var(-0.75)

        # Check output computation (weighted sum + bias)

        # test without activation
        neuron.activation = None
        self.assertEqual(neuron([Var(1), Var(-0.5), Var(1.5)]).data, 1.75)  # 1*0 + (-0.5)*1 + 1.5*2 + 1*(-0.75)
        self.assertEqual(neuron([Var(1), Var(-3), Var(1.5)]).data, -0.75)  # 1*0 + (-3)*1 + 1.5*2 + 1*(-0.75)

        neuron.activation = 'linear'
        self.assertEqual(neuron([Var(1), Var(-0.5), Var(1.5)]).data, 1.75)  # 1*0 + (-0.5)*1 + 1.5*2 + 1*(-0.75)
        self.assertEqual(neuron([Var(1), Var(-3), Var(1.5)]).data, -0.75)  # 1*0 + (-3)*1 + 1.5*2 + 1*(-0.75)

        # test with relu activation function
        neuron.activation = 'relu'
        self.assertEqual(neuron([Var(1), Var(-0.5), Var(1.5)]).data, 1.75)
        self.assertEqual(neuron([Var(1), Var(-3), Var(1.5)]).data, 0)

        # test with tanh activation function
        neuron.activation = 'tanh'
        self.assertEqual(round(neuron([Var(1), Var(-0.5), Var(1.5)]).data), 1.0)
        self.assertEqual(round(neuron([Var(1), Var(-3), Var(1.5)]).data), -1.0)

        # test with sigmoid activation function
        neuron.activation = 'sigmoid'
        self.assertEqual(round(neuron([Var(1), Var(-0.5), Var(1.5)]).data), 1.0)
        self.assertEqual(round(neuron([Var(1), Var(-3), Var(1.5)]).data), 0.0)

    def test_layer(self):
        layer = Layer(in_features=3, out_features=2)
        self.assertEqual(len(layer.neurons), 2)
        self.assertEqual(len(layer.parameters()), 8)  # 3*2 + 2

        for neuron in layer.neurons:
            self.assertIsNone(neuron.activation)

        layer_with_activation = Layer(in_features=3, out_features=2, activation='relu')
        self.assertEqual(len(layer_with_activation.neurons), 2)
        self.assertEqual(len(layer_with_activation.parameters()), 8)  # 3*2 + 2
        for neuron in layer_with_activation.neurons:
            self.assertIs(neuron.activation, 'relu')

    def test_mlp(self):
        with self.assertRaises(AssertionError):
            incorrect_model_object_1 = MLP(in_features=2, layers=[3, 1], activations='relu')
            incorrect_model_object_2 = MLP(in_features=2, layers=[3, 1], activations=['relu'])

        model = MLP(in_features=2, layers=[3, 1], activations=['relu', 'linear'])  # 1 hidden layer with 3 neurons

        self.assertEqual(len(model.layers), 2)
        self.assertEqual(len(model.parameters()), 13)  # 2*3+3 + 3*1+1

        for layer, activation in zip(model.layers, ['relu', 'linear']):
            for neuron in layer.neurons:
                self.assertIs(neuron.activation, activation)


if __name__ == "__main__":
    unittest.main()



import unittest

from picograd.engine import Var
from picograd.optim import SGD, Adam


class TestOptimizer(unittest.TestCase):

    def test_sgd(self):
        params = [Var(3), Var(-1), Var(0.5)]
        sgd = SGD(parameters=params, lr=0.1)

        # Simulate gradient update (normally done through backprop)
        for i, p in enumerate(params):
            p.grad = i

        sgd.step()
        self.assertEqual(params[0].data, 3)  # 3 - 0.1 * 0
        self.assertEqual(params[1].data, -1.1)  # -1 - 0.1 * 1
        self.assertEqual(params[2].data, 0.3)  # 0.5 - 0.1 * 2

        sgd.zero_grad()
        for _, p in enumerate(params):
            self.assertEqual(p.grad, 0)

    def test_sgd_with_momentum_and_nesterov(self):
        params = [Var(3), Var(-1), Var(0.5)]
        sgd = SGD(parameters=params, lr=0.1, momentum=0.9, nesterov=True)

        # Simulate gradient update (normally done through backprop)
        for i, p in enumerate(params):
            p.grad = i

        sgd.step()
        self.assertEqual(params[0].data, 3)  # 3 - (1 + 0.9) * 0.1 * 0
        self.assertEqual(params[1].data, -1.19)  # -1 - (1 + 0.9) * 0.1 * 1
        self.assertEqual(params[2].data, 0.12)  # 0.5 - (1 + 0.9) * 0.1 * 2

        sgd.zero_grad()
        for _, p in enumerate(params):
            self.assertEqual(p.grad, 0)

    def test_adam(self):
        params = [Var(3), Var(-1), Var(0.5)]
        sgd = Adam(parameters=params, lr=1e-3)

        # Simulate gradient update (normally done through backprop)
        for i, p in enumerate(params):
            p.grad = i

        sgd.step()
        self.assertAlmostEqual(params[0].data, 3, 2)
        self.assertAlmostEqual(params[1].data, -1, 2)
        self.assertAlmostEqual(params[2].data, 0.5, 2)

        sgd.zero_grad()
        for _, p in enumerate(params):
            self.assertEqual(p.grad, 0)


if __name__ == "__main__":
    unittest.main()



import random
from abc import ABC, abstractmethod
from picograd.engine import Var

from typing import List, Optional, Union


class Module(ABC):
    def __init(self) -> None:
        pass

    @abstractmethod
    def parameters(self) -> List[Var]:
        raise NotImplementedError

    def zero_grad(self) -> None:
        for p in self.parameters():
            p.grad = 0.0


class Neuron(Module):
    """A single neuron"""

    def __init__(self, in_features: int, activation: Optional[str] = None):
        self.w = [Var(random.uniform(-1, 1)) for _ in range(in_features)]
        self.b = Var(random.uniform(-1, 1))
        self.activation = activation

    def __call__(self, x):
        # w * x + b
        out = sum([w_i * x_i for w_i, x_i in zip(self.w, x)], self.b)
        if self.activation is None or self.activation == 'linear':
            return out
        elif self.activation == 'relu':
            return out.relu()
        elif self.activation == 'tanh':
            return out.tanh()
        elif self.activation == 'sigmoid':
            return out.sigmoid()
        raise NotImplementedError(
            f"Unexpected activation argument ('relu', 'tanh' and 'sigmoid' available). Got {self.activation}.")

    def parameters(self) -> List[Var]:
        return self.w + [self.b]

    def __repr__(self) -> str:
        return f"Neuron({len(self.w)}, {self.activation if self.activation is not None else 'linear'})"


class Layer(Module):
    """A layer of neurons"""

    def __init__(self, in_features: int, out_features: int, activation: Optional[str] = None):
        self.neurons = [Neuron(in_features, activation) for _ in range(out_features)]

    def __call__(self, x: List[Var]) -> List[Var]:
        outs = [n(x) for n in self.neurons]
        return outs  # outs[0] if len(outs) == 1 else outs

    def parameters(self) -> List[Var]:
        return [p for n in self.neurons for p in n.parameters()]

    def __repr__(self) -> str:
        return f"Layer of [{', '.join(str(n) for n in self.neurons)}]"


class MLP(Module):
    """ A Multi-layer Perceptron """

    def __init__(self, in_features: int, layers: List[int], activations: List[str]):
        sizes = [in_features] + layers
        assert len(activations) != 0, "Please provide activation for layers. Available -> 'relu', 'tanh', 'sigmoid', 'linear'"
        assert len(activations) == len(layers), "length of activations does not match the length of layers"
        self.layers = [Layer(in_features=sizes[i], out_features=sizes[i + 1], activation=activations[i])
                       for i in range(len(layers))]

    def __call__(self, x: List[Var]) -> Union[List[Var], Var]:
        for layer in self.layers:
            x = layer(x)
        return x

    def parameters(self) -> List[Var]:
        return [p for layer in self.layers for p in layer.parameters()]

    def __repr__(self) -> str:
        return f"MLP of [{', '.join(str(layer) for layer in self.layers)}]"



import unittest

from picograd.engine import Var
from picograd.metrics import mean_squared_error, binary_accuracy


class TestMetrics(unittest.TestCase):
    def test_mse(self):
        y_true = [Var(3), Var(-0.5), Var(2), Var(7)]
        y_pred = [Var(2.5), Var(0.0), Var(2), Var(8)]
        error = mean_squared_error(y_true, y_pred)
        self.assertAlmostEqual(error.data, 0.375, 3)

    def test_binary_accuracy(self):
        y_true = [Var(0), Var(1), Var(2), Var(3)]
        y_pred = [Var(0), Var(2), Var(1), Var(3)]

        acc = binary_accuracy(y_true, y_pred)
        self.assertEqual(acc, 0.5)

        acc = binary_accuracy(y_true, y_true)
        self.assertEqual(acc, 1.0)


if __name__ == "__main__":
    unittest.main()



from picograd.nn import Module
from picograd.optim import Optimizer
from picograd.data import BatchIterator

from typing import Callable, Dict, List

# Used to record training history for metrics
History = Dict[str, List[float]]


class Trainer:
    """Encapsulates the model training loop"""

    def __init__(self, model: Module, optimizer: Optimizer, loss: Callable, acc_metric: Callable) -> None:
        self.model = model
        self.optimizer = optimizer
        self.loss = loss
        self.acc_metric = acc_metric

    def fit(self, data_iterator: BatchIterator, num_epochs: int = 500, verbose: bool = False) -> History:
        """Fits the model to the data"""

        history: History = {"loss": [], "acc": []}
        epoch_loss = 0
        epoch_acc = 0
        epoch_y_true = []
        epoch_y_pred = []
        for epoch in range(num_epochs):
            # Reset the gradients of model parameters
            self.optimizer.zero_grad()
            # Reset epoch data
            epoch_loss = 0
            epoch_y_true = []
            epoch_y_pred = []

            for batch in data_iterator():
                # Forward pass
                # outputs = [self.model(mini_batch_input) for mini_batch_input in batch.inputs]
                outputs = list(map(self.model, batch.inputs))

                # Loss computation
                batch_y_pred = [item for sublist in outputs for item in sublist]
                batch_loss = self.loss(batch.targets, batch_y_pred)
                epoch_loss += batch_loss.data

                # Store batch predictions and ground truth for computing epoch metrics
                epoch_y_pred.extend(batch_y_pred)
                epoch_y_true.extend(batch.targets)

                # Backprop and gradient descent
                batch_loss.backward()
                self.optimizer.step()

            # Accuracy computation for epoch
            epoch_acc = self.acc_metric(epoch_y_true, epoch_y_pred)

            # Record training history
            history["loss"].append(epoch_loss)
            history["acc"].append(epoch_acc)

            if verbose:
                print(
                    f"Epoch [{epoch + 1}/{num_epochs}], "
                    f"loss: {epoch_loss:.6f}, "
                    f"accuracy: {epoch_acc * 100:.2f}%"
                )

        return history



from abc import ABC, abstractmethod
from graphviz import Digraph

from picograd.engine import Var


class ComputationalGraphViz(ABC):
    """
        ComputationalGraphViz is the building block object for quick
        visualization of a computational graph. Picograd dynamically
        builds a computational graph and keeps track of the operations
        made while building the computational graph.
        Performing depth-first-search over the computational graph, ComputationalGraphViz
        is able to build a trace from a root node. Usually the root node
        is simply the final result of an operation. For MLP, it is usually the loss function.
    """

    def __init__(self):
        self._nodes = set()
        self._edges = set()

    def create_graph(self, root: Var, rankdir: str = 'LR') -> Digraph:
        """
        Builds the computational graph and displays it
        :param root: root node of type Var
        :param rankdir: TB (top to bottom graph) | LR (left to right)
        :return: Digraph
        """

        self._build_trace(root)
        graph = self._build_graph(rankdir=rankdir)
        return graph

    def _build_trace(self, node: Var) -> None:
        """Performs a recursive depth-first search over the computational graph."""
        if node not in self._nodes:
            self._nodes.add(node)
            for child in node.children:
                self._edges.add((child, node))
                self._build_trace(child)

    @abstractmethod
    def _build_graph(self, rankdir: str) -> Digraph:
        raise NotImplementedError("build_graph() must be implemented for subclasses!")


class ForwardGraphViz(ComputationalGraphViz):

    def __init__(self):
        super(ForwardGraphViz, self).__init__()

    def _build_graph(self, rankdir: str = 'LR') -> Digraph:
        """
        Builds the forward graph
        :param rankdir: TB (top to bottom graph) | LR (left to right)
        :return: Digraph
        """

        assert rankdir in ['LR', 'TB'], f"Unexpected rankdir argument (TB, LR available). Got {rankdir}."
        graph = Digraph(format='png', graph_attr={'rankdir': rankdir})

        for n in self._nodes:
            uid = str(id(n))
            name = n.label if n.label != "" else (n.op + '_res' if n.op else n.label)
            # for any value in the graph, create a rectangular ('record') node for it
            graph.node(name=uid, label=f"{name} | data: {n.data:.4f} | grad: {n.grad:.4f}", shape='record')
            if n.op:
                graph.node(name=uid + n.op, label=n.op)
                graph.edge(uid + n.op, uid)

        for n1, n2 in self._edges:
            # connect n1 to the op node of n2
            if n2.op:
                graph.edge(str(id(n1)), str(id(n2)) + n2.op)
            else:
                graph.edge(str(id(n1)), str(id(n2)))

        return graph



"""
Autograd engine implementing reverse-mode auto-differentiation, aka backpropagation.
"""

import math
from typing import Union, Tuple, List, Set, Callable

FloatInt = Union[float, int]


class Var:
    """ stores a single scalar value and its gradient """

    def __init__(self, data: FloatInt, children: Tuple["Var", ...] = (), op: str = "",
                 label: str = "") -> None:
        self.data: FloatInt = data
        self.grad: float = 0.0

        # Internal variables used for autograd graph construction
        self._backward: Callable = lambda: None
        self._prev: Set[Var] = set(children)
        self._op: str = op  # The operation that produced this node, for graphviz / debugging / etc
        self._label: str = label

    @property
    def children(self):
        return self._prev

    @property
    def op(self):
        return self._op

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, label_name: str):
        self._label = label_name

    def __repr__(self):
        if len(self._label) != 0:
            return f"Var(data={self.data}, label={self.label})"
        else:
            return f"Var(data={self.data})"

    def __add__(self, other: Union["Var", FloatInt]) -> "Var":
        other = other if isinstance(other, Var) else Var(other)
        out = Var(self.data + other.data, children=(self, other), op="+")

        def _backward() -> None:
            self.grad += 1.0 * out.grad  # local_grad * global_grad -> chain rule
            other.grad += 1.0 * out.grad

        out._backward = _backward

        return out

    def __neg__(self) -> "Var":  # -self
        return self * -1

    def __radd__(self, other: Union["Var", FloatInt]) -> "Var":  # other + self
        return self + other

    def __sub__(self, other: Union["Var", FloatInt]) -> "Var":  # self - other
        return self + (-other)

    def __rsub__(self, other: Union["Var", FloatInt]) -> "Var":  # other - self
        return -self + other

    def __mul__(self, other: Union["Var", FloatInt]) -> "Var":
        other = other if isinstance(other, Var) else Var(other)
        out = Var(self.data * other.data, children=(self, other), op='*')

        def _backward() -> None:
            self.grad += other.data * out.grad
            other.grad += self.data * out.grad

        out._backward = _backward

        return out

    def __rmul__(self, other: Union["Var", FloatInt]) -> "Var":  # other * self
        return self * other

    def __pow__(self, other: FloatInt) -> "Var":
        assert isinstance(other, (int, float)), "only supporting int/flot powers for now"
        out = Var(self.data ** other, children=(self,), op=f'**{other}')

        def _backward() -> None:
            self.grad += other * (self.data ** (other - 1)) * out.grad

        out._backward = _backward

        return out

    def __truediv__(self, other: Union["Var", FloatInt]) -> "Var":  # self / other
        return self * other ** -1

    def __rtruediv__(self, other: Union["Var", FloatInt]) -> "Var":  # other / self
        return other * self ** -1

    def exp(self) -> "Var":
        """Compute exp()"""

        x = self.data
        out = Var(math.exp(x), children=(self,), op='exp')

        def _backward():
            self.grad += out.data * out.grad

        out._backward = _backward

        return out

    def tanh(self) -> "Var":
        """Compute tanh()"""

        x = self.data
        t = (math.exp(2 * x) - 1) / (math.exp(2 * x) + 1)
        out = Var(t, children=(self,), op='tanh')

        def _backward() -> None:
            self.grad += (1 - t ** 2) * out.grad

        out._backward = _backward

        return out

    def relu(self) -> "Var":
        """Compute ReLU"""

        out = Var(0 if self.data < 0 else self.data, children=(self,), op='ReLU')

        def _backward() -> None:
            self.grad += (out.data > 0) * out.grad

        out._backward = _backward

        return out

    def sigmoid(self) -> "Var":
        """Compute sigmoid()"""

        x = self.data
        s = 1 / (1 + math.exp(-x))
        out = Var(s, children=(self,), op='sigmoid')

        def _backward() -> None:
            self.grad += (1 - s) * out.grad

        out._backward = _backward

        return out

    def backward(self) -> None:
        """Compute gradients through backpropagation"""

        # Topological order of all the children in the graph from left to right edges
        topo: List[Var] = []
        visited: Set[Var] = set()

        def build_topo(node: Var) -> None:
            if node not in visited:
                visited.add(node)
                for child in node._prev:
                    build_topo(child)
                topo.append(node)

        build_topo(self)

        # Go one variable at a time and apply the chain rule to get its gradient
        self.grad = 1.0
        for node in reversed(topo):
            node._backward()


if __name__ =="__main__":
    from picograd.graph_viz import ForwardGraphViz

    graph_builder = ForwardGraphViz()

    x = Var(1.0, label='x')
    y = (x * 2 + 1).relu()
    y.label = 'y'
    y.backward()

    graph_builder.create_graph(y).view()


import random
from picograd.engine import Var

from typing import NamedTuple, Iterator, List

Batch = NamedTuple("Batch", [("inputs", List[List[Var]]), ("targets", List[Var])])


class BatchIterator:
    """Iterates on data by batches"""

    def __init__(self, inputs: List[List[Var]], targets: List[Var], batch_size: int = 32, shuffle: bool = True) -> None:
        self.inputs = inputs
        self.targets = targets
        self.batch_size = batch_size
        self.shuffle = shuffle

    def __call__(self) -> Iterator[Batch]:
        starts = list(range(0, len(self.inputs), self.batch_size))
        if self.shuffle:
            random.shuffle(starts)

        for start in starts:
            end = start + self.batch_size
            batch_inputs = self.inputs[start:end]
            batch_targets = self.targets[start:end]
            yield Batch(inputs=batch_inputs, targets=batch_targets)



"""A setuptools based setup module.
See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

from setuptools import setup, find_packages

with open("README.md") as f:
    README = f.read()

setup(
    name='picograd',
    version='1.0.1',
    description='A lightweight machine learning framework',
    url='https://github.com/shubhamwagh/picograd',
    long_description=README,
    long_description_content_type='text/markdown',
    license="MIT",
    platforms=['Ubuntu 20.04', 'Ubuntu 21.04', 'Windows'],
    author='Shubham Wagh',
    author_email='shubhamwagh48@gmail.com',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        'Programming Language :: Python :: 3 :: Only',
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: Developers",
    ],
    keywords='picograd, autograd, backprop, nn, graph',
    packages=find_packages(),
    python_requires='>=3.6, <4',
    data_files=[('misc', ['misc/moon_mlp.png', 'misc/simple_graph.png'])],
    install_requires=[],
)



import unittest
from collections import Counter

from picograd.engine import Var
from picograd.data import BatchIterator


class TestDataUtils(unittest.TestCase):

    def test_batch_iterator(self):
        # Generate lists of 8 scalars with same values
        DATASET_SIZE = 8
        inputs = [[Var(x)] for x in range(DATASET_SIZE)]
        targets = [Var(x) for x in range(DATASET_SIZE)]

        batch_size = 3
        data_iterator = BatchIterator(inputs, targets, batch_size=batch_size)

        batch_sizes = []
        for batch in data_iterator():
            # Check lengths of batches
            self.assertEqual(len(batch.inputs), len(batch.targets))
            self.assertLessEqual(len(batch.inputs), batch_size)
            self.assertLessEqual(len(batch.targets), batch_size)

            # Store batch size (same for inputs and targets)
            batch_sizes.append(len(batch.inputs))

        # Check expected sizes of batches (exact order can vary)
        self.assertEqual(Counter(batch_sizes), Counter([3, 3, 2]))


if __name__ == "__main__":
    unittest.main()



import unittest

from picograd.engine import Var
from picograd.nn import MLP
from picograd.optim import SGD
from picograd.metrics import mean_squared_error, binary_accuracy
from picograd.data import BatchIterator
from picograd.trainer import Trainer, History


class TestTrainer(unittest.TestCase):
    def test_trainer(self):
        # dataset for AND logical function
        x_train = [
            list(map(Var, x)) for x in [[0, 0], [0, 1], [1, 0], [1, 1]]
        ]
        y_train = [Var(0), Var(0), Var(0), Var(1)]

        model = MLP(in_features=2, layers=[1], activations=['linear'])  # Logistic regression
        optimizer = SGD(model.parameters(), lr=0.1)
        data_iterator = BatchIterator(x_train, y_train)

        num_epochs = 50
        trainer = Trainer(model, optimizer, loss=mean_squared_error, acc_metric=binary_accuracy)
        history: History = trainer.fit(data_iterator, num_epochs=num_epochs, verbose=False)

        # Training metrics are recorded for each epoch
        self.assertEqual(len(history["loss"]), len(history["acc"]))
        self.assertEqual(len(history["loss"]), num_epochs)
        self.assertEqual(len(history["acc"]), num_epochs)

        # Access final values for metrics
        loss = history["loss"][-1]
        acc = history["acc"][-1]

        self.assertLess(loss, 0.1)
        self.assertEqual(acc, 1.0)  # 100%


if __name__ == "__main__":
    unittest.main()



from picograd.engine import Var

from typing import List


def mean_squared_error(y_true: List[Var], y_pred: List[Var]) -> Var:
    """MSE loss"""
    assert len(y_true) == len(y_pred)
    total_squared_error = sum([(y_true_i - y_pred_i) ** 2 for y_true_i, y_pred_i in zip(y_true, y_pred)], 0.0)
    n_total = max(len(y_true), 1)
    return total_squared_error / n_total


def binary_accuracy(y_true: List[Var], y_pred: List[Var]) -> float:
    """Binary accuracy"""
    assert len(y_true) == len(y_pred)
    n_exact = sum([y_true_i.data == round(y_pred_i.data) for y_true_i, y_pred_i in zip(y_true, y_pred)], 0)
    n_total = max(len(y_true), 1)
    return n_exact / n_total
